--[[
* Update date: 27.08.2021
* Version: 1.09
* Category: Items
* Description: Render selected pre-glued items that have active MIDI takes and source track instrument to new take without FX
* Author: Yaunick
* Contact: b.yanushevich@gmail.com
* Donation: https://paypal.me/yaunick?locale.x=ru_RU
--]]

  --------------------------------------------------------------------------------------------

    tail_for_item = 0   --- sec
    user_inputs = true
    
    bypass_active_preFX_envelopes = true
      warn_about_active_preFX_env = true
      
    warn_about_item_and_take_volume = true
      
    warn_about_overlapping_MIDI_items_and_that_not_all_items_on_the_tracks_are_selected = true   --- without warning, the script will run faster
      overlapping_detect_sensitivity = 10    --- millisec (for ignoring overlaps associated with MIDI PPQ)
      
    after_rendering_set_the_instrument_to = 0   --- 0 = leave active
                                                --- 1 = set bypass
                                                --- 2 = set offline
                 
  --------------------------------------------------------------------------------------------
  
  function bla() end
  function nothing() reaper.defer(bla) end
  
  if (not tonumber(tail_for_item) or tail_for_item < 0)
  or (user_inputs ~= true and user_inputs ~= false)
  or (bypass_active_preFX_envelopes ~= true and bypass_active_preFX_envelopes ~= false)
  or (warn_about_active_preFX_env ~= true and warn_about_active_preFX_env ~= false)
  or (warn_about_item_and_take_volume ~= true and warn_about_item_and_take_volume ~= false)
  or (
    warn_about_overlapping_MIDI_items_and_that_not_all_items_on_the_tracks_are_selected ~= true
    and warn_about_overlapping_MIDI_items_and_that_not_all_items_on_the_tracks_are_selected ~= false
  )
  or (not tonumber(overlapping_detect_sensitivity) or overlapping_detect_sensitivity < 0)
  or (after_rendering_set_the_instrument_to ~= 0 and after_rendering_set_the_instrument_to ~= 1 and after_rendering_set_the_instrument_to ~= 2)
  then
    reaper.MB('Incorrect settings at the beginning of the scripts', 'Error', 0)
    nothing() return
  end
  
  local test_SWS = reaper.CF_EnumerateActions
  if not test_SWS then
    reaper.MB('Please install or update SWS extension', 'Error', 0) nothing() return
  end
  
  if reaper.CountSelectedMediaItems(0) == 0 then
    reaper.MB("No items. Please select an item", "Error",0)
    nothing() return
  end
  
  local t_tracks = {}
  local t_prefx_env = {}
  local t_track_itms = {}
  local msg = false
  local msg_2 = false
  local overlap = false
  local test_instr = true
  
  for i=0, reaper.CountSelectedMediaItems(0)-1 do
    local item_1 = reaper.GetSelectedMediaItem(0,i)
    
    local track_1 = reaper.GetMediaItemTrack(item_1)
    if test_instr == true then
      if reaper.TrackFX_GetInstrument(track_1) == -1 then
        reaper.MB("Please select items on tracks with online instruments", "Error", 0)
        nothing() return
      end
      for h=1, reaper.TrackFX_GetCount(track_1) do
        local _, buf = reaper.TrackFX_GetFXName(track_1, h-1, '')
        if buf:sub(0,6) == "VSTi: " and h-1 ~= reaper.TrackFX_GetInstrument(track_1) then
          reaper.MB("Please select items on tracks with only one instrument", "Error", 0)
          nothing() return
        end
      end
      test_instr = false
    end
    
    local take = reaper.GetActiveTake(item_1)
    if reaper.TakeIsMIDI(take) == false then
      reaper.MB("Please select only items with active MIDI takes", "Error", 0)
      nothing() return
    end
    if reaper.TakeFX_GetCount(take) > 0 then
      reaper.MB('Please select only takes with empty take FX chain', 'Error', 0)
      nothing() return
    end
    if reaper.CountTakeEnvelopes(take) > 0 then
      reaper.MB('Please select only takes without active take envelopes', 'Error', 0)
      nothing() return
    end
    if warn_about_item_and_take_volume == true then
      if msg_2 == false then
        if reaper.GetMediaItemInfo_Value(item_1, 'D_VOL') ~= 1 
        or reaper.GetMediaItemTakeInfo_Value(take, 'D_VOL') ~= 1 then
          local retval =
          reaper.MB('You have volume values on the selected items that change the velocity of MIDI takes. ' ..
          'After rendering, the velocity values will be applied and the new velocity values will be displayed inside the MIDI take. ' ..
          '\n\nAre you sure you want to continue?','Warning',1)
          if retval == 2 then
            nothing() return
          end
          msg_2 = true
        end
      end
    end
    if i == 0 then
      t_tracks[#t_tracks+1] = track_1
    end
    local item_2 = reaper.GetSelectedMediaItem(0,i+1)
    if item_2 then
      local track_2 = reaper.GetMediaItemTrack(item_2)
      if track_1 ~= track_2 then
        t_tracks[#t_tracks+1] = track_2
        test_instr = true
      end
    end
  end
  
  if warn_about_overlapping_MIDI_items_and_that_not_all_items_on_the_tracks_are_selected == true then
    for i=1, #t_tracks do
      for f=0, reaper.CountTrackMediaItems(t_tracks[i])-1 do
        local track_itm_1 = reaper.GetTrackMediaItem(t_tracks[i], f)
        if after_rendering_set_the_instrument_to == 1 or after_rendering_set_the_instrument_to == 2 then
          if reaper.TakeIsMIDI(reaper.GetActiveTake(track_itm_1)) == true 
          and reaper.IsMediaItemSelected(track_itm_1) == false then
            local retval = 
            reaper.MB("You haven't selected all active MIDI takes on the tracks. Because the instrument after rendering will cease to be active, " ..
            'the remaining active MIDI takes will not sound. ' ..
            '\n\nAre you sure you want to continue?', 
            'Warning', 1)
            if retval == 2 then
              nothing() return
            else
              break
            end
          end
        else
          local track_itm_2 = reaper.GetTrackMediaItem(t_tracks[i], f+1)
          if track_itm_2 then
            local track_itm_end_1 = reaper.GetMediaItemInfo_Value(track_itm_1, 'D_POSITION') + 
            reaper.GetMediaItemInfo_Value(track_itm_1, 'D_LENGTH')
            local track_itm_start_2 = reaper.GetMediaItemInfo_Value(track_itm_2, 'D_POSITION')
            if track_itm_end_1 - (overlapping_detect_sensitivity/1000) > track_itm_start_2 then
              overlap = true
            else
              overlap = false
            end
            if overlap == true then
              if reaper.TakeIsMIDI(reaper.GetActiveTake(track_itm_1)) == true and reaper.TakeIsMIDI(reaper.GetActiveTake(track_itm_2)) == true
              then
                if (reaper.IsMediaItemSelected(track_itm_1) == true and reaper.IsMediaItemSelected(track_itm_2) == false)
                or (reaper.IsMediaItemSelected(track_itm_1) == false and reaper.IsMediaItemSelected(track_itm_2) == true)
                then
                  local retval = reaper.MB('Among the overlapping sets of active MIDI takes, there are selected and unselected takes at the same time. ' ..
                  'After rendering, If MIDI and AUDIO overlapping takes are mixed, the final sound may change. \n\nAre you sure you want to continue?',
                  'Warning', 1)
                  if retval == 2 then
                    nothing() return
                  else
                    break
                  end
                end
              end
            end
          end
        end
      end
    end
  end
  
  if bypass_active_preFX_envelopes == true then
    for i=1, #t_tracks do
      for g=1, reaper.CountTrackEnvelopes(t_tracks[i]) do
        local tr_envelo = reaper.GetTrackEnvelope(t_tracks[i], g-1)
        local tr_envelo_br = reaper.BR_EnvAlloc(tr_envelo, true)
        local active, visible, armed, inLane, laneHeight, 
        defaultShape, _, _, _, type_env, faderScaling, automationItemsOptions = reaper.BR_EnvGetProperties(tr_envelo_br)
        if type_env == 1 or type_env == 3 or type_env == 5 then
          if active == true then
            if warn_about_active_preFX_env == true then
              if msg == false then
                local retval = reaper.MB("You have active Volume Pre-FX or Pan Pre-FX or Width Pre-FX envelopes! " .. 
                "\n\nAfter rendering, the script should be set the bypass state of these envelopes, " ..
                "because they will be applied to the new audio take and the final sound may change. " ..
                "\n\nAre you sure you want to continue?", "Warning", 1)
                if retval == 2 then
                  nothing() return
                end
                msg = true
              end
            end
            t_prefx_env[#t_prefx_env+1] = {tr_envelo, visible, armed, inLane, laneHeight, defaultShape, faderScaling, automationItemsOptions}
          end
        end
        reaper.BR_EnvFree(tr_envelo_br, true)
      end
    end
  end
  
  if user_inputs == true then
    ::START::
    local retval, value = reaper.GetUserInputs("Render selected MIDI takes to new take", 1, 
    "Set the tail for new item:,extrawidth=40",tail_for_item)
    if not retval then
      nothing() return
    end
    if not tonumber(value) or tonumber(value) < 0 then
      reaper.MB("Incorrect value. Set the number of tail", "Error" , 0)
      goto START
    end
    tail_for_item = tonumber(value)
  end
  
  reaper.Undo_BeginBlock()
  reaper.PreventUIRefresh(1)
  
  local loopnewit_save = reaper.SNM_GetIntConfigVar('loopnewitems',0)
  local change_setting = false
  if loopnewit_save&32 ~= 32 then
    local loopnewit = loopnewit_save~32
    reaper.SNM_SetIntConfigVar('loopnewitems',loopnewit)
    change_setting = true
  end
  
  local apply_takes = 40209
  local t_fx, t_env, t_instr_byp = {}, {}, {}
  
  reaper.Main_OnCommand(40362,0) --Item: Glue items ignoring time selection
  
  reaper.ApplyNudge(0, 0, 3, 1, tail_for_item, 0, 0)
  
  for i=1, #t_tracks do
    for s=1, reaper.TrackFX_GetCount(t_tracks[i]) do
      local num_par = reaper.TrackFX_GetNumParams(t_tracks[i], s-1)
      local env_fx = reaper.GetFXEnvelope(t_tracks[i], s-1, num_par-2, false)  --- get native Bypass FX envelope
      if env_fx ~= nil and s-1 == reaper.TrackFX_GetInstrument(t_tracks[i]) 
      and after_rendering_set_the_instrument_to == 1 then
        t_instr_byp[#t_instr_byp+1] = {s-1, t_tracks[i], num_par-2}
      elseif env_fx ~= nil and s-1 > reaper.TrackFX_GetInstrument(t_tracks[i]) then
        local br_envelope = reaper.BR_EnvAlloc(env_fx, true)
        
        local active, visible, armed, inLane, laneHeight, 
        defaultShape, _, _, _, _, faderScaling, automationItemsOptions = reaper.BR_EnvGetProperties(br_envelope)
        
        t_env[#t_env+1] = {
        env_fx, active, visible, armed, inLane, laneHeight, defaultShape, faderScaling, automationItemsOptions
        }
        
        reaper.BR_EnvSetProperties(br_envelope, false, visible, armed, 
        inLane, laneHeight, defaultShape, faderScaling, automationItemsOptions)
        
        reaper.BR_EnvFree(br_envelope, true)
      end
    end
    
    for j=1, reaper.TrackFX_GetCount(t_tracks[i]) do
      if j-1 > reaper.TrackFX_GetInstrument(t_tracks[i]) then
        t_fx[#t_fx+1] = { j-1, t_tracks[i], reaper.TrackFX_GetEnabled( t_tracks[i], j-1) }
        reaper.TrackFX_SetEnabled( t_tracks[i], j-1, false)
      end
    end
  end
   
  reaper.Main_OnCommand(apply_takes,0)
  
  for s=1, #t_env do
    local br_envelope = reaper.BR_EnvAlloc( t_env[s][1], true)
    
    reaper.BR_EnvSetProperties(br_envelope, t_env[s][2], t_env[s][3], t_env[s][4], 
    t_env[s][5], t_env[s][6], t_env[s][7], t_env[s][8], t_env[s][9])
    
    reaper.BR_EnvFree(br_envelope, true)
  end
  
  for j=1, #t_fx do
    reaper.TrackFX_SetEnabled( t_fx[j][2], t_fx[j][1], t_fx[j][3])
  end

  for i=1, #t_instr_byp do
    local env_byp = reaper.GetFXEnvelope(t_instr_byp[i][2], t_instr_byp[i][1], t_instr_byp[i][3], false)
    local br_envelope = reaper.BR_EnvAlloc(env_byp, true)
    local active, visible, armed, inLane, laneHeight, 
    defaultShape, _, _, _, _, faderScaling, automationItemsOptions = reaper.BR_EnvGetProperties(br_envelope)
            
    reaper.BR_EnvSetProperties(br_envelope, false, visible, armed, 
    inLane, laneHeight, defaultShape, faderScaling, automationItemsOptions)
            
    reaper.BR_EnvFree(br_envelope, true)
  end
  
  for i=1, #t_prefx_env do
    local br_envelope = reaper.BR_EnvAlloc(t_prefx_env[i][1], true)
    reaper.BR_EnvSetProperties(br_envelope, false, t_prefx_env[i][2], t_prefx_env[i][3], t_prefx_env[i][4],
    t_prefx_env[i][5], t_prefx_env[i][6], t_prefx_env[i][7], t_prefx_env[i][8])
    reaper.BR_EnvFree(br_envelope, true)
  end

  if after_rendering_set_the_instrument_to > 0 then
    for i=1, #t_tracks do
      local tr_instr = reaper.TrackFX_GetInstrument(t_tracks[i])
      if after_rendering_set_the_instrument_to == 1 then
        reaper.TrackFX_SetEnabled(t_tracks[i], tr_instr, 0)
      else
        reaper.TrackFX_SetOffline(t_tracks[i], tr_instr, 1)
      end
    end
  end
  
  if change_setting == true then
    reaper.SNM_SetIntConfigVar('loopnewitems', loopnewit_save)
  end
  
  reaper.Undo_EndBlock(
  'Render selected pre-glued items that have active MIDI takes and source track instrument to new take without FX',
  -1)
  reaper.PreventUIRefresh(-1)
